# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
This module provides a _DomainProviderManagedProjectDefinition_ class
that can be used to represent managed projects definitions.
"""

from typing import Any, Callable, Dict, List, NoReturn, Optional, Set

from zabel.commons.exceptions import ApiError
from zabel.commons.interfaces import ManagedProjectDefinition
from zabel.commons.utils import (
    ensure_in,
    ensure_nonemptystring,
    ensure_noneorinstance,
)


########################################################################
########################################################################

## managed project definitions


class DomainProviderManagedProjectDefinition(ManagedProjectDefinition):
    """Managed Project Definition with Domains and Providers.

    It extends _ManagedProjectDefinition_.

    A _DomainProviderManagedProject_ contains a set of _domains_ (e.g.
    tools or services) that themselves contain a set of _providers_
    (e.g. teams).

    Each provider contains _members_, split in _categories_, and a set
    of additional fields.

    By default there are three member categories ('admins', 'readers'
    and 'users') and one additional field ('options').

    This wrapper offers a set of helpers managing members and additional
    fields in domain providers.

    ## Usage

    If an implementation wants to make use of the offered helpers, it
    will have to declare the list of its domains in the `DOMAINS`
    attribute.

    Each domain will be able to contain any number of providers.  Those
    providers then can have members ('admins', 'users' and 'readers' by
    default, as specified by the `MEMBER_CATEGORIES` attribute) and
    additional fields ('options' by default, as specified by the
    `OTHER_CATEGORIES` attribute).

    ## Attributes

    | Attribute           | Description                                |
    | ------------------- | ------------------------------------------ |
    | `DOMAINS`           | A dictionary.  The keys are the internal
                            names (the entry in the MPD) and the values
                            are the public names (what will be used in
                            the accessors method names).<br/>
                            The internal names will be keys in the
                            managed project definition dictionaries and
                            public names will be used in accessors
                            method names.<br/>
                            Internal and public names should not start
                            with an underscore.                        |
    | `MEMBER_CATEGORIES` | A dictionary.  The keys are the members
                            categories names, and the values are their
                            types (_dict_ being the only supported
                            value).<br/>
                            The 'member' category name is reserved.    |
    | `OTHER_CATEGORIES`  | A dictionary.  The keys are the 'other'
                            categories names, and the values are their
                            types (_list_ being the only supported
                            value).                                    |

    ## Accessors

    Accessors methods for domains are automatically generated if not
    redefined in the children class.

    For each domain, the accessors methods are (assuming the default
    categories for providers):

    - `list_{public_name}s()`
    - `create_{public_name}(provider)`

    - `list_{public_name}s_members()`
    - `delete_{public_name}s_member()`

    - `list_{public_name}_admins(provider)`
    - `add_{public_name}_admin(provider, name, service='*')`
    - `remove_{public_name}_admin(provider, name, service=None)`
    - `list_{public_name}_users(provider)`
    - `add_{public_name}_user(provider, name, service='*')`
    - `remove_{public_name}_user(provider, name, service=None)`
    - `list_{public_name}_readers(provider)`
    - `add_{public_name}_reader(provider, name, service='*')`
    - `remove_{public_name}_reader(provider, name, service=None)`
    - `list_{public_name}_options(provider)`
    - `add_{public_name}_option(provider, item)`
    - `remove_{public_name}_option(provider, item)`

    In addition to those domain accessors, the following methods are
    defined:

    - `list_members()`
    - `remove_member(item)`
    - `list_domain_provider_{category}s(domain, provider)`
    - `add_domain_provider_{category}(domain, provider, item, service='*')`
    - `remove_domain_provider_{category}(domain, provider, item, service=None)`
    """

    DOMAINS: Dict[str, str] = {}
    MEMBER_CATEGORIES: Dict[str, type] = {
        'admin': list,
        'user': list,
        'reader': list,
    }
    OTHER_CATEGORIES: Dict[str, type] = {'option': list}

    def __init__(self) -> None:
        """Create a new managed project."""
        super().__init__()

        if 'member' in self.MEMBER_CATEGORIES:
            raise ApiError('Key \'member\' is reserved in MEMBER_CATEGORIES')

        for category in self.MEMBER_CATEGORIES:
            self._expose_domain_provider_membercategoryaccessor(category)
        for category in self.OTHER_CATEGORIES:
            self._expose_domain_provider_othercategoryaccessor(category)
        self['spec'] = {}
        for domain in self.DOMAINS:
            self['spec'][domain] = {}
            self._expose_accessors(domain)

    # Domains
    #

    def is_domain_publicname(self, domain: str) -> bool:
        """Return whether a domain is a domain public name."""
        return domain in self.DOMAINS.values()

    def get_domain_publicname(self, internal: str) -> str:
        """Return domain's public name."""
        return self.DOMAINS[internal]

    def get_domain_internalname(self, domain: str) -> str:
        """Return domain's internal name."""
        return [d for d in self.DOMAINS if self.DOMAINS[d] == domain][0]

    # Members
    #

    def list_members(self) -> Set[str]:
        """Return the set of domain provider members."""
        members: Set[str] = set()
        for domain in self.DOMAINS:
            members = members.union(self.list_domain_members(domain))
        return members

    def remove_member(self, member: str) -> None:
        """Remove member from all domain providers."""
        for domain in self.DOMAINS:
            self.remove_domain_member(domain, member)

    # Domains Members Helpers
    #
    # list_domain_members
    # remove_domain_member

    def list_domain_members(self, domain: str) -> Set[str]:
        """Return a set of declared domain members.

        `domain` is the domain internalname.

        # Required parameters

        - domain: a non-empty string

        # Returned value

        A set.  Each item in the set is a string.
        """
        members: Set[str] = set()
        for provider in self['spec'][domain]:
            for category in self.MEMBER_CATEGORIES:
                members = members.union(
                    m['account'] for m in provider[f'{category}s']
                )
        return members

    def remove_domain_member(self, domain: str, member: str) -> None:
        """Remove a member from all domain providers.

        `domain` is the domain internal name.

        # Required parameters

        - domain: a non-empty string
        - member: a non-empty string

        # Returned value

        None.
        """
        for provider in self['spec'][domain]:
            for category in self.MEMBER_CATEGORIES:
                members = provider[f'{category}s']
                for item in [m for m in members if m['account'] == member]:
                    members.remove(item)

    # Domains Providers Helpers
    #
    # list_domain_providers
    # create_domain_provider
    # delete_domain_provider

    def list_domain_providers(self, domain: str) -> Set[str]:
        """Return a set of declared domain providers.

        # Required parameters

        - `domain`: a non-empty string

        # Returned value

        A set.  Each item in the set is a string.
        """
        if not domain.startswith('__'):
            ensure_in('domain', self.DOMAINS)

        return set(provider['name'] for provider in self['spec'][domain])

    def create_domain_provider(
        self, domain: str, provider: str, **fields: Any
    ) -> None:
        """Add a provider to a domain.

        # Required parameters

        - `domain`: a non-empty string
        - `provider`: a non-empty string

        # Optional parameters

        - `fields`: a dictionary or None

        # Raised exceptions

        If `provider` is already declared in `domain`, an _ApiError_
        exception is raised.
        """
        ensure_nonemptystring('provider')
        ensure_noneorinstance('fields', dict)
        if not domain.startswith('__'):
            ensure_in('domain', self.DOMAINS)

        if provider in self.list_domain_providers(domain):
            raise ApiError(
                'Provider %s already declared in domain %s.'
                % (provider, domain)
            )

        _definition = {'name': provider}
        self['spec'][domain].append(_definition)

        for category in self.MEMBER_CATEGORIES:
            _definition[f'{category}s'] = self.MEMBER_CATEGORIES[category]()
        for category in self.OTHER_CATEGORIES:
            _definition[f'{category}s'] = self.OTHER_CATEGORIES[category]()
        if fields:
            for field in fields:
                _definition[field] = fields[field]

    def delete_domain_provider(self, domain: str, provider: str) -> None:
        """Remove provider.

        # Required parameters

        - `domain`: a non-empty string
        - `provider`: a non-empty string

        # Raised exceptions

        If `provider` is not declared in `domain`, an _ApiError_
        exception is raised.
        """
        ensure_nonemptystring('provider')
        if not domain.startswith('__'):
            ensure_in('domain', self.DOMAINS)

        if provider not in self.list_domain_providers(domain):
            raise ApiError(
                'Provider %s not declared in domain %s.' % (provider, domain)
            )

        for item in [p for p in self['spec'][domain] if p['name'] == provider]:
            self['spec'][domain].remove(item)

    def list_domain_provider_category_members(
        self, domain: str, provider: str, category: str
    ) -> List[Dict[str, Any]]:
        """Return specified members for domain provider.

        `domain` is the domain internal name.

        # Required parameters

        - `domain`: a non-empty string
        - `provider`: a non-empty string
        - `category`: a non-empty string

        # Returned value

        A list of dictionaries.

        # Raised exceptions

        Raises an _ApiError_ exception if `provider`, `domain`, or
        `category` are not known.
        """
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    return list(p[category])
            else:
                self._raise_keyerror(domain, provider, category)
        except KeyError:
            self._raise_keyerror(domain, provider, category)

    # private helpers

    def _raise_keyerror(
        self, domain: str, provider: str, item: str, key: Optional[str] = None
    ) -> NoReturn:
        if domain not in self['spec']:
            raise ApiError('Domain %s not known.' % domain)
        if provider not in self['spec'][domain]:
            raise ApiError(
                'Provider %s not known in domain %s.' % (provider, domain)
            )
        if item not in self['spec'][domain][provider]:
            raise ApiError(
                'Item %s not known in provider %s for domain %s.'
                % (item, provider, domain)
            )
        raise ApiError(
            '%s %s not known in provider %s for domain %s.'
            % (item.title(), key, provider, domain)
        )

    def _add_category_member_option(
        self,
        domain: str,
        provider: str,
        category: str,
        member: str,
        option: str,
    ) -> None:
        """Add option to member, or create member with specified option."""
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    members = p[category]
                    break
            else:
                self._raise_keyerror(domain, provider, category)
            for m in members:
                if m['account'] == member:
                    if option in m['options']:
                        raise ApiError(
                            '%s %s already set to %s for provider %s.'
                            % (category, member, option, provider)
                        )
                    m['options'].append(option)
                    break
            else:
                members.append({'account': member, 'options': [option]})
        except KeyError:
            self._raise_keyerror(domain, provider, category)

    def _remove_category_member_option(
        self,
        domain: str,
        provider: str,
        category: str,
        member: str,
        option: Optional[str],
    ) -> None:
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    members = p[category]
                    break
            else:
                self._raise_keyerror(domain, provider, category, member)
            for m in members:
                if m['account'] == member:
                    if option is not None:
                        if option in m['options']:
                            m['options'].remove(option)
                        else:
                            raise ApiError(
                                '%s %s not set to %s for provider %s.'
                                % (category, member, option, provider)
                            )
                        if option is None or not m['options']:
                            members.remove(m)
                        break
            else:
                self._raise_keyerror(domain, provider, category, member)
        except KeyError:
            self._raise_keyerror(domain, provider, category, member)

    def _list_domain_provider_category_items(
        self, domain: str, provider: str, category: str
    ) -> List[str]:
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    return list(p[category])
            else:
                self._raise_keyerror(domain, provider, category)
        except KeyError:
            self._raise_keyerror(domain, provider, category)

    def _add_category_item(
        self, domain: str, provider: str, category: str, item: str
    ) -> None:
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    if item in p[category]:
                        raise ApiError(
                            '%s %s already declared for provider %s'
                            % (category, item, provider)
                        )
                    p[category].append(item)
                    break
            else:
                self._raise_keyerror(domain, provider, category)
        except KeyError:
            self._raise_keyerror(domain, provider, category)

    def _remove_category_item(
        self, domain: str, provider: str, category: str, item: str
    ) -> None:
        ensure_nonemptystring('domain')
        ensure_nonemptystring('provider')
        ensure_nonemptystring('category')

        try:
            for p in self['spec'][domain]:
                if p['name'] == provider:
                    if item not in p['category']:
                        raise ApiError(
                            '%s %s not declared for provider %s'
                            % (category, item, provider)
                        )
                    p[category].remove(item)
                    break
            else:
                self._raise_keyerror(domain, provider, category)
        except KeyError:
            self._raise_keyerror(domain, provider, category)

    # accessors generators

    def _create_method(self, name: str, body: Callable[..., Any]) -> None:
        if not hasattr(self, name):
            self.__setattr__(name, body)

    def _expose_member_category_accessors(
        self, domain: str, name: str, category: str
    ) -> None:
        """Create undefined accessors for domain member categories.

        Will attempt to create the following methods:

        - `list_{domain}_{category}s(provider)`
        - `add_{domain>_{category}(provider, member, service='*')`
        - `remove_{domain}_{category}(provider, member, service=None)`

        If a method already exists, it is left as-is.
        """
        cat = '%ss' % category
        self._create_method(
            'list_%s_%ss' % (name, category),
            lambda provider: self.list_domain_provider_category_members(
                domain, provider, cat
            ),
        )
        self._create_method(
            'add_%s_%s' % (name, category),
            lambda provider, member, service='*': self._add_category_member_option(
                domain, provider, cat, member, service
            ),
        )
        self._create_method(
            'remove_%s_%s' % (name, category),
            lambda provider, member, service=None: self._remove_category_member_option(
                domain, provider, cat, member, service
            ),
        )

    def _expose_other_category_accessors(
        self, domain: str, name: str, category: str
    ) -> None:
        """Create undefined accessors for domain 'other' categories.

        Will attempt to create the following methods:

        - `list_{domain}_{category}s(provider)`
        - `add_{domain}_{category}(provider, item)`
        - `remove_{domain}_{category}(provider, item)`

        If a method already exists, it is left as-is.
        """
        cat = '%ss' % category
        self._create_method(
            'list_%s_%ss' % (name, category),
            lambda provider: self._list_domain_provider_category_items(
                domain, provider, cat
            ),
        )
        self._create_method(
            'add_%s_%s' % (name, category),
            lambda provider, option: self._add_category_item(
                domain, provider, cat, option
            ),
        )
        self._create_method(
            'remove_%s_%s' % (name, category),
            lambda provider, option: self._remove_category_item(
                domain, provider, cat, option
            ),
        )

    def _expose_accessors(self, domain: str) -> None:
        """Create accessors for domain, if not already defined.

        Will attempt to create the following methods:

        - `list_{domain}s()`
        - `create_{domain}(provider)`
        - `delete_{domain}(provider)`
        - `list_{domain}s_members()`
        - `remove_{domain}s_member()`

        Will also attempt to create the category accessors.

        If a method already exists, it is left as-is.
        """
        name = self.DOMAINS[domain]

        # providers
        self._create_method(
            'list_%ss' % name, lambda: self.list_domain_providers(domain)
        )
        self._create_method(
            'create_%s' % name,
            lambda provider: self.create_domain_provider(domain, provider),
        )
        self._create_method(
            'delete_%s' % name,
            lambda provider: self.delete_domain_provider(domain, provider),
        )

        self._create_method(
            'list_%ss_members' % name, lambda: self.list_domain_members(domain)
        )

        self._create_method(
            'remove_%ss_member' % name,
            lambda member: self.remove_domain_member(domain, member),
        )

        # providers members
        for category in self.MEMBER_CATEGORIES:
            self._expose_member_category_accessors(domain, name, category)

        for category in self.OTHER_CATEGORIES:
            self._expose_other_category_accessors(domain, name, category)

    def _expose_domain_provider_membercategoryaccessor(
        self, category: str
    ) -> None:
        """Create domain_provider accessors for member category.

        Will attempt to create the following methods:

        - `list_domain_provider_{category}s(domain, provider)`
        - `add_domain_provider_{category}(domain, provider, member,
            service='*')`
        - `remove_domain_provider_{category}(domain, provider, member,
            service=None)`

        If a method already exists, it is left as-is.
        """
        cat = '%ss' % category
        self._create_method(
            'list_domain_provider_%ss' % category,
            lambda domain, provider: self.list_domain_provider_category_members(
                domain, provider, cat
            ),
        )
        self._create_method(
            'add_domain_provider_%s' % category,
            lambda domain, provider, member, service='*': self._add_category_member_option(
                domain, provider, cat, member, service
            ),
        )
        self._create_method(
            'remove_domain_provider_%s' % category,
            lambda domain, provider, member, service=None: self._remove_category_member_option(
                domain, provider, cat, member, service
            ),
        )

    def _expose_domain_provider_othercategoryaccessor(
        self, category: str
    ) -> None:
        """Create domain_provider accessors for other category.

        Will attempt to create the following methods:

        - `list_domain_provider_{category}s(domain, provider)`
        - `add_domain_provider_{category}(domain, provider, item)`
        - `remove_domain_provider_{category}(domain, provider, item)`

        If a method already exists, it is left as-is.
        """
        cat = '%ss' % category
        self._create_method(
            'list_domain_provider_%ss' % category,
            lambda domain, provider: self._list_domain_provider_category_items(
                domain, provider, cat
            ),
        )
        self._create_method(
            'add_domain_provider_%s' % category,
            lambda domain, provider, member: self._add_category_item(
                domain, provider, cat, member
            ),
        )
        self._create_method(
            'remove_domain_provider_%s' % category,
            lambda domain, provider, member: self._remove_category_item(
                domain, provider, cat, member
            ),
        )



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
This module implements the specifics for the Zabel managed projects.

It provides a _DomainProviderManagedProjectDefinition_ subclass that can
be used to model many projects definitions.

It also provides a _ManagedProjectDefinitionManager_ class.
"""

__all__ = [
    'DomainProviderManagedProjectDefinition',
    'ManagedProjectDefinitionManager',
]

from .domainprovidermpd import DomainProviderManagedProjectDefinition

from .mpdmanager import ManagedProjectDefinitionManager



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
**amac** storage wrappers.

Implements some internal classes:

- #Storag: an interface for storage classes
- #ObjectStorage: an interface for storage classes handling one object
- #CollectionStorage: an interface for storage classes handling a
    collection of objects.

- #AWSS3Storage: an abstract class providing AWS S3 helpers
- #AWSS3Object: a Storage class for JSON files stored on a S3
- #AWSS3Bucket: a Storage class for handling S3 buckets

- #ManagedDict: a simple class making use of an #ObjectStorage
    delegate.
"""

from typing import Any, Dict, Iterator, List, Optional

import json

from zabel.commons.exceptions import ApiError
from zabel.commons.utils import ensure_instance, ensure_nonemptystring


########################################################################
## Storage Interfaces
##
## Storage
## ObjectStorage
## CollectionStorage


class Storage:
    """Abstract base storage wrapper.

    Declares the minimal set of features storage implementation must
    provide:

    - constructor (`__init__`)

    The constructor will take one parameter, a dictionary.  Its content
    is implementation-dependent.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, configuration: Dict[str, Any]) -> None:
        """Initialize a new storage object.

        # Required parameters

        - `configuration`: a dictionary

        The `configuration` dictionary content is
        implementation-dependent.
        """
        ensure_instance('configuration', dict)

        self.configuration = configuration


class ObjectStorage(Storage):
    """Abstract object storage wrapper.

    Declares the minimal set of features storage implementation must
    provide:

    - `read`
    - `create`
    - `update`
    - `delete`

    The data format is unspecified, and is implementation-dependent.
    """

    def read(self) -> Any:
        """Return object from storage."""
        raise NotImplementedError

    def create(self, data: Any) -> None:
        """Create object in storage."""
        raise NotImplementedError

    def update(self, data: Any) -> None:
        """Update object in storage."""
        raise NotImplementedError

    def delete(self) -> None:
        """Remove object from storage."""
        raise NotImplementedError


class CollectionStorage(Storage):
    """Abstract collection storage wrapper.

    Declares the minimal set of features storage implementation must
    provide:

    - `list`
    - `read`
    - `create`
    - `update`
    - `delete`

    The data format is unspecified, and is implementation-dependent.
    """

    def list(self) -> List[Dict[str, Any]]:
        """Return list of items in storage.

        # Returned value

        A list. Each element in the list is a dictionary with at least
        a `key` entry:

        - `key`: a string

        It may contain a `last_modified` entry too, if appropriate:

        - `last_modified`: a datetime object
        """
        raise NotImplementedError

    def read(self, key: str) -> Any:
        """Return specified item from storage."""
        raise NotImplementedError

    def create(self, key: str, data: Any) -> None:
        """Create specified item in storage."""
        raise NotImplementedError

    def update(self, key: str, data: Any) -> None:
        """Update specified item in storage."""
        raise NotImplementedError

    def delete(self, key: str) -> None:
        """Remove specified item from storage."""
        raise NotImplementedError


########################################################################
## AWS Storage


class AWSS3Storage(Storage):
    """A base Storage abstract class for AWS S3 classes.

    Assumes the configuration dictionary contains the following entries:

    ```python
    {
        'storage': {
            'bucket': 'a string'
        }
    }
    ```
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, configuration: Dict[str, Any]) -> None:
        """Initialize a new AWSS3Storage object."""
        super().__init__(configuration)
        from boto3.resources.factory import ServiceResource

        self._s3: Optional[ServiceResource] = None

    # Helpers

    def _get_bucket(self) -> Any:
        """Return the bucket."""
        if self._s3 is None:
            import boto3

            self._s3 = boto3.resource('s3')
        if self._s3 is None:
            raise ApiError('AWS S3 service resource is None.')
        return self._s3.Bucket(self.configuration['storage']['bucket'])

    def _get_object(self, name: str) -> Any:
        """Return the object."""
        return self._get_bucket().Object(name)


class AWSS3Bucket(AWSS3Storage, CollectionStorage):
    """Simple AWS S3 Storage class.

    This class handles objects stored in a S3 bucket.

    It uses the following configuration entry:

    ```python
    {
        'storage': {
            'bucket': 'a string'
        }
    }
    ```

    It must be an existing bucket name.

    Objects are stored as JSON objects.
    """

    def list(self) -> List[Dict[str, Any]]:
        return [
            {'key': item.key, 'last_modified': item.last_modified}
            for item in self._get_bucket().objects.all()
        ]

    def read(self, key: str) -> Dict[str, Any]:
        ensure_nonemptystring('key')

        obj = self._get_object(key)
        return json.loads(obj.get()['Body'].read().decode('utf-8'))  # type: ignore

    def create(self, key: str, data: Dict[str, Any]) -> None:
        ensure_nonemptystring('key')
        ensure_instance('data', dict)

        if self._has(key):
            raise ApiError('Object %s already exist.' % key)

        obj = self._get_object(key)
        obj.put(Body=bytes(json.dumps(data, default=str), 'utf-8'))

    def update(self, key: str, data: Dict[str, Any]) -> None:
        ensure_nonemptystring('key')
        ensure_instance('data', dict)

        if not self._has(key):
            raise ApiError('Object %s does not exist.' % key)

        obj = self._get_object(key)
        obj.put(Body=bytes(json.dumps(data, default=str), 'utf-8'))

    def _has(self, key: str) -> bool:
        _exists = True
        import botocore

        try:
            self._get_object(key).load()
        except botocore.exceptions.ClientError as error:
            if error.response['Error']['Code'] == '404':
                _exists = False
            else:
                raise

        return _exists


class AWSS3Object(AWSS3Storage, ObjectStorage):
    """Simple AWS S3 Object Storage class.

    This class handles objects stored in a S3 bucket.

    It enhance the platform definition with the following entries:

    ```python
    {
        'storage': {
            'bucket': 'a string',
            'filename': 'a string'
        }
    }
    ```

    It must be an existing bucket name, and file name is expected to
    be an existing and valid JSON object.

    The bucket may contain other objects, they will be ignored.
    """

    def read(self) -> Dict[str, Any]:
        """Return stored object as JSON."""
        obj = self._get_object(self.configuration['storage']['filename'])
        return json.loads(obj.get()['Body'].read().decode('utf-8'))  # type: ignore

    def update(self, data: Dict[str, Any]) -> None:
        """Update stored object."""
        obj = self._get_object(self.configuration['storage']['filename'])
        obj.put(Body=bytes(json.dumps(data, default=str), 'utf-8'))


########################################################################
## Useful Helpers


class ManagedDict(Dict[str, Any]):
    """Simple wrapper for dictionaries.

    It expects an _ObjectStorage_ object defined in its configuration
    data.

    Entries starting with an underscore ('_') are 'hidden', in that they
    are not returned by the `__iter__` method and `__contains__` does
    not see them (but they can be used to store values and are
    persisted).

    In other words, assuming `foo` is a managed dictionary:

    ```python
    >>> foo['bar'] = 123
    >>> foo['_bar'] = 456

    >>> foo['bar']              # => 123
    >>> foo['_bar']             # => 456
    >>> 'bar' in foo            # => True
    >>> '_bar' in foo           # => False
    >>> keys = [k for k in foo] # => ['bar']
    >>> sorted(foo)             # => ['bar']

    >>> # but
    >>> len(foo)                # => 2
    >>> foo.keys()              # => dict_keys(['bar', '_bar'])
    >>> foo.items()             # => dict_items([('bar', 123),
    >>>                         #                ('_bar', 456)])
    >>> foo.values()            # => dict_values([123, 456])
    ```
    """

    def __init__(self, configuration: Dict[str, Any]) -> None:
        """Initialize a managed dict.

        The initial content of the managed dictionary is read according
        to `configuration`.

        Changes are allowed during the life of the managed dictionary.
        Use the `persist` method to save them.

        # Required parameters

        - `configuration`: a dictionary

        `configuration` is a dictionary with the following entry:

        - `storage`: a dictionary

        `storage` is a dictionary with at least the following entry:

        - `type`: a class subclassing __ObjectStorage__

        It may also contain additional entries, depending on its `type`
        class.
        """
        super().__init__()
        self._storage = configuration['storage']['type'](configuration)
        source = self._storage.read()
        for key in source:
            self[key] = source[key]

    def __iter__(self) -> Iterator[str]:
        for k in super().__iter__():
            if not k.startswith('_'):
                yield k

    def __contains__(self, key: object) -> bool:
        return (
            isinstance(key, str)
            and super().__contains__(key)
            and not key.startswith('_')
        )

    def persist(self) -> None:
        """Persist the dictionary."""
        self._storage.update(self)



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
The high-level **zabel** library.

It provides an abstraction for _platforms_, which are collections of
_services_.

## Services (Abstract Base Classes)

### Realm Abstract Class

**work in progress, not ready yet**

| Abstract&nbsp;Class | Description                                    |
| ------------------- | ---------------------------------------------- |
| _Realm_             | A collection of platforms and domains.  A realm
                        handles users and managed projects, and
                        implements the #::Service interface.           |

### Platform Abstract Class

| Abstract&nbsp;Class | Description                                    |
| ------------------- | ---------------------------------------------- |
| #::Platform         | A service that manages a collection of
                        managedservices and utilities services, with an
                        associated set of properties and datasources.
                        A platform handles users and managed projects,
                        and implements the #::Service interface.       |

## Properties Helper Classes

### Managed accounts

| Helper&nbsp;Class        | Description                               |
| ------------------------ | ----------------------------------------- |
| #::ManagedAccount        | An abstract class that represents a
                             minimal managed account.                  |
| #::ManagedAccountManager | An abstract class that handles collections
                             of managed accounts.                      |

### Managed projects

| Helper&nbsp;Class                        | Description               |
| ---------------------------------------- | ------------------------- |
| #::DomainProviderManagedProjectDefinition| An abstract class that
                                             extends the
                                             _ManagedProjectDefinition_
                                             class to cover common
                                             managed project definition
                                             needs.                    |
| #::ManagedProjectDefinitionManager       | An abstract class that
                                             handles collections of
                                             managed project
                                             definitions.              |

## Datasources Helper Classes

| Helper&nbsp;Class   | Description                                    |
| ------------------- | ---------------------------------------------- |
| #::Storage          | An interface for storage classes.              |
| #::ObjectStorage    | An interface for storage classes handling one
                        object.                                        |
| #::CollectionStorage| An interface for storage classes handling a
                        collection of objects.                         |
| #::AWSS3Storage     | An abstract class providing AWS S3 helpers.    |
| #::AWSS3Object      | A _Storage_ class for JSON files stored on a
                        S3 bucket.                                     |
| #::AWSS3Bucket      | A _Storage_ class for handling S3 buckets.     |
| #::ManagedDict      | A simple class making use of an _ObjectStorage_
                        delegate, providing a 'persistent' dictionary. |
"""

__all__ = [
    'ManagedAccount',
    'ManagedAccountManager',
    'DomainProviderManagedProjectDefinition',
    'ManagedProjectDefinitionManager',
    'Service',
    'Utility',
    'ManagedService',
    'Platform',
]


from zabel.commons.interfaces import (
    Service,
    Utility,
    ManagedService,
)

from .services import Platform
from .accounts import (
    ManagedAccount,
    ManagedAccountManager,
)
from .projects import (
    DomainProviderManagedProjectDefinition,
    ManagedProjectDefinitionManager,
)



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
The **zabel.fabric**  _Platform_ abstract class.
"""

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
)

from zabel.commons.exceptions import ApiError
from zabel.commons.interfaces import (
    Service,
    ManagedService,
    ManagedProjectDefinition,
    ManagedAccount,
)
from zabel.commons.utils import (
    api_call,
    ensure_instance,
    ensure_nonemptystring,
    ensure_noneorinstance,
)

from ..accounts import ManagedAccountManager
from ..projects import ManagedProjectDefinitionManager


########################################################################
## Service


class Platform(Service):
    """A _Platform_ service.

    A platform is a collection of _services_ with associated _members_,
    _properties_, and _managed projects_.

    This class has no public constructor.  You have to use the
    #::amac.make_platform() factory method to instantiate a _Platform_
    object.

    # Services

    Services are of two kinds: _utilities_, that are not hosting managed
    projects, and _managed services_, that are hosting managed projects.

    Utilities are for example directories or external services (AWS,
     Azure, ...).

    Managed services are typically tools (Artifactory, Jira, ...).

    If no credentials are provided for a service, the service will not
    be instantiated (except if it is defined as being an _anonymous_
    service).

    # Members

    A platform has associated members (members of the managed services
    that are associated to the platform).

    Members have _canonical_ IDs, which are platform-specific.

    Each service must implement the necessary translation methods, to
    and from those canonical IDs from and to their internal IDs.

    Canonical IDs are strings.  Internal IDs are service-specific
    (but typically either strings or integers).

    # Properties

    Properties are data objects attached to a platform.  Those
    properties can be literals (say, a text description) or 'live'
    objects (instances of a class, for example #::Manager or
    #::ManagedDict).

    Properties have a name and a value.  They are singletons.

    Each platform has at least two associated properties,
    `managedprojectdefinitions`, which is an object implementing the
    #::ManagedProjectDefinitionManager interface and `managedaccounts`,
    which is an object implementing the #::ManagedAccountManager
    interface.

    # Managed projects

    Managed projects each have a definition.  Those definitions are
    objects implementing the #::ManagedProjectDefinition interface.

    Managed projects can be 'pushed' to the platform managed services.

    # Managed accounts

    Managed accounts represent the managed projects members, and may
    possibly represent former managed projects members, as well as
    other members (for example, platform administrators, ...).

    Managed accounts can be queried or disabled.

    ## Platform attributes

    In addition to the methods it provides, 3 attributes are exposed:

    | Attribute     | Description                                      |
    | ------------- | ------------------------------------------------ |
    | `definition`  | A dictionary, the platform definition            |
    | `credentials` | A #::Credentials object                          |
    | `services`    | A dictionary where keys are service names and
                      values are instances of the specified services   |
    """

    def __init__(
        self, _: str, definition: Dict[str, Any], credentials: Credentials
    ) -> None:
        self.definition = definition
        self.credentials = credentials
        self.services: Dict[str, Service] = {}
        self._mpdmanager: Optional[ManagedProjectDefinitionManager] = None
        self._mamanager: Optional[ManagedAccountManager] = None

        for name, service in definition['services'].items():
            if service.get('anonymous', False) or credentials.contains(name):
                self.services[name] = _make_service(name, service, credentials)
        for service in self.services.values():
            service.platform = self

    def __str__(self) -> str:
        return f'{self.__class__.__name__}: {self.definition["name"]!r}'

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.definition["name"]!r}>'

    ####################################################################
    # platform members
    #
    # list_members
    # get_member

    @api_call
    def list_members(self) -> Dict[str, Dict[str, Any]]:
        """Return the members of the service.

        Member of the platform are those of the managed services it
        contains.

        # Returned value

        A dictionary.  The keys are the canonical IDs and the values are
        dictionaries.

        Those dictionaries have one entry per platform service in which
        the member is known.

        Entries are the name of the service, and values are the service
        member representation.
        """

        def _add_member(
            members: Dict[str, Dict[str, Any]],
            name: str,
            member: str,
            user: Any,
        ) -> None:
            if member not in members:
                members[member] = dict()
            members[member][name] = user

        members: Dict[str, Dict[str, Any]] = dict()
        for name, service in self.services.items():
            if not isinstance(service, ManagedService):
                continue
            service_members = service.list_members()
            for member in service_members:
                _add_member(members, name, member, service_members[member])

        return members

    @api_call
    def get_member(self, member_id: str) -> Dict[str, Any]:
        """Return the user details.

        # Required parameters

        - member_id: a string

        `member_id` is the canonical member ID.

        # Returned value

        A dictionary.  See #list_members() for more information on
        its content.

        # Raised exceptions

        An _ApiError_ exception is raised if `member_id` is not a member
        of the platform.
        """
        ensure_nonemptystring('member_id')

        member = dict()
        for name, service in self.services.items():
            if isinstance(service, ManagedService):
                try:
                    member[name] = service.get_member(member_id)
                except ApiError:
                    pass

        if not member:
            raise ApiError('Member %s not found on platform' % member_id)

        return member

    ####################################################################
    # platform renderer
    #

    def render(self, target: Any = None) -> None:
        """Render the platform object for the specified target."""
        raise NotImplementedError

    ####################################################################
    # platform properties
    #

    _properties: Dict[str, Any] = {}

    def get_property(self, name: str) -> Any:
        """Return platform property.

        # Required parameters

        - name: a string

        # Returned value

        A data object, which is either the literal value as stored
        in the platform definition or an instance of a class specified
        in the `type` entry of the literal.

        This class is expected to have a constructor with one parameter,
        which will be the dictionary associated with the property.

        For example, if the 'properties' section in the platform
        definition contains the following:

        ```python
        {
            'foo': [1, 2, 3],
            'bar': {
                'type': Something,
                'arg': [1, 2, 3]
            }
        }
        ```

        `get_property('foo')` will return `[1, 2, 3]`.

        `get_property('bar')` will return the result of the following
        call:

        ```python
        Something({'type': Something, 'arg': [1, 2, 3]})
        ```

        Properties are singletons.
        """
        if name not in self._properties:
            what = self.definition['properties'][name]
            if isinstance(what, dict) and 'type' in what:
                self._properties[name] = what['type'](what)
            else:
                self._properties[name] = what
        return self._properties[name]

    ####################################################################
    # platform managed accounts
    #
    # get_managedaccount
    # create_managedaccount
    # update_managedaccount

    @api_call
    def get_managedaccount(self, canonical_id: str) -> ManagedAccount:
        """Return managed account details.

        Please refer to the concrete managed account used for more
        information on its content.

        # Required parameters

        - canonical_id: a non-empty string

        # Return value

        An instance of _ManagedAccount_.
        """
        ensure_nonemptystring('canonical_id')

        return self._get_mamanager().get_managedaccount(canonical_id)

    @api_call
    def create_managedaccount(
        self, canonical_id: str, account: ManagedAccount
    ) -> None:
        """Create new managed account.

        # Required parameters

        - canonical_id: a string
        - account: a #::ManagedAccount object

        # Raised exceptions

        `canonical_id` must not be the ID of an existing managed
        account.  If it is, an _ApiError_ exception will be raised.
        """
        ensure_nonemptystring('canonical_id')
        ensure_instance('account', ManagedAccount)

        self._get_mamanager().create_managedaccount(canonical_id, account)

    @api_call
    def update_managedaccount(
        self, canonical_id: str, account: ManagedAccount
    ) -> None:
        """Update managed account

        # Required parameters

        - canonical_id: a string
        - account: a #::ManagedAccount object

        # Raised exceptions

        If no existing managed account with the provided ID exists,
        an _ApiError_ exception is raised.
        """
        ensure_nonemptystring('canonical_id')
        ensure_instance('account', ManagedAccount)

        self._get_mamanager().update_managedaccount(canonical_id, account)

    ####################################################################
    # platform managed projects
    #
    # list_managedprojects
    # get_managedproject
    # create_managedproject
    # update_managedproject
    # push_managedproject
    # render_managedproject

    # reading/writing projects

    @api_call
    def list_managedprojects(self) -> List[Dict[str, Any]]:
        """Return list of managed projects on platform.

        # Returned value

        Each item in the list is a dictionary with the following
        entries:

        - project_id: a string
        - last_modified: a _datetime.datetime_ object
        """
        return self._get_mpdmanager().list_managedprojects()

    @api_call
    def get_managedproject(self, project_id: str) -> ManagedProjectDefinition:
        """Return managed project details.

        Please refer to the concrete managed project definition used for
        more information on its content.

        # Required parameters

        - project_id: a non-empty string

        # Returned value

        An instance of #::ManagedProjectDefinition.
        """
        ensure_nonemptystring('project_id')

        return self._get_mpdmanager().get_managedproject(project_id)

    @api_call
    def create_managedproject(
        self, project_id: str, project: ManagedProjectDefinition
    ) -> None:
        """Create new managed project.

        # Required parameters

        - project_id: a string
        - project: a dictionary

        # Raised exceptions

        `project_id` must not be the ID of an existing managed project.
        If it is, an _ApiError_ exception will be raised.
        """
        ensure_nonemptystring('project_id')
        ensure_instance('project', dict)

        self._get_mpdmanager().create_managedproject(project_id, project)

    @api_call
    def update_managedproject(
        self, project_id: str, project: ManagedProjectDefinition
    ) -> None:
        """Update managed project definition.

        # Required parameters

        - project_id: a string
        - project: a dictionary

        # Raised exceptions

        If no existing managed project with the provided ID exists,
        an _ApiError_ exception is raised.
        """
        ensure_nonemptystring('project_id')
        ensure_instance('project', dict)

        self._get_mpdmanager().update_managedproject(project_id, project)

    def render_managedproject(
        self, project_id: str, target: Any = None
    ) -> Any:
        """Render the managed project on the specified target."""
        raise NotImplementedError

    # pushing projects

    def get_push_strategy(
        self, project: ManagedProjectDefinition, context: Mapping[str, Any]
    ) -> Iterable[Callable[[ManagedProjectDefinition], None]]:
        """Return the push strategy.

        !!! important
            Subclasses must implement this method if the platform allows
            for managed project pushes.

        This method is called by #push_managedproject() to get the _push
        strategy_ for the given project and context.

        A push strategy is an iterable that returns _push steps_.

        A push step is a function that takes one argument, the project
        definition, and returns no value but may raise an exception.

        Steps are called in order by #push_managedproject().

        If a step raises an exception, the possible remaining steps are
        ignored, and the rollback strategy will be queried and executed
        if the exception is an _ApiError_ exception.

        If all steps complete successfully, #push_managedproject() will
        return True.

        # Required parameters

        - project: a _ManagedProjectDefinition_
        - context: a dictionary

        # Returned value

        A possibly empty list of push steps (callables taking a
        #::ManagedProjectDefinition as their only parameter).
        """
        raise NotImplementedError

    def get_rollback_strategy(
        self,
        project: ManagedProjectDefinition,
        trace: Iterable[Callable[[ManagedProjectDefinition], None]],
        ex: ApiError,
    ) -> Iterable[Callable[[ManagedProjectDefinition], None]]:
        """Return the failed push rollback strategy.

        This function is called by #push_managedproject() if an
        _ApiError_ exception occurs while executing the push strategy.

        By default, it returns a _rollback strategy_ that simply
        re-raise the exception.

        A platform may choose to refine this rollback strategy,
        possibly allowing for 'clean' failures.

        A rollback strategy is an iterable that returns _rollback
        steps_.

        A rollback step is a function that takes one argument, the
        project definition (a #::ManagedProjectDefinition).

        If the returned strategy is empty, nothing will be attempted,
        and the #push_managedproject() function will simply return
        False.

        If the returned strategy is not empty, steps will be called in
        order.  If a step raises an exception, it will be propagated
        (and the possible remaining steps will be ignored).

        If all steps complete successfully (i.e., they don't raise an
        exception), the #push_managedproject() function will return
        False with no further ado.

        # Required parameters

        - project: a #::ManagedProjectDefinition
        - trace: a list of performed actions, the last one having
          raised an _ApiError_ exception
        - ex: the raised exception

        If `trace` is the empty list, the exception was raised while
        acquiring the push strategy.

        # Returned value

        A possibly empty list of rollback steps (callables taking a
        single parameter, a #::ManagedProjectDefinition).
        """
        # pylint: disable=unused-argument,no-self-use
        def _raise(_: ManagedProjectDefinition) -> None:
            raise ex

        return [_raise]

    @api_call
    def push_managedproject(
        self,
        project_id: str,
        context: Optional[MutableMapping[str, Any]] = None,
    ) -> bool:
        """Push (aka publish) managed project on platform.

        This method queries the platform push strategy by calling
        #get_push_strategy() and execute it.

        If an _ApiError_ exception occurs while executing the push
        strategy, a roll-back strategy is queried by calling
        #get_rollback_strategy() and is then performed.

        # Required parameters

        - project_id: a non-empty string

        # Optional parameters

        - context: a dictionary (None by default)

        If `context` is provided, a transient `_trace` entry will be
        added (or, if already there, will be reinitialized) and will
        contain the resulting execution _trace_ (the collection of
        values passed to info, warning, and debug).  The `_trace` value
        is a list of tuples: (stamp, severity, message).  The list is
        ordered (most recent entry last).

        # Returned value

        A boolean.  True if the managed project is successfully pushed,
        False if a successful rollback was performed.

        Raises an exception otherwise.
        """
        ensure_nonemptystring('project_id')
        ensure_noneorinstance('context', dict)

        project = self.get_managedproject(project_id)
        if context is None:
            context = {}

        trace = []
        try:
            for step in self.get_push_strategy(project, context):
                trace.append(step)
                step(project)
        except ApiError as ex:
            context['_trace'] = project['_transient']['trace']
            for step in self.get_rollback_strategy(project, trace, ex):
                step(project)
            return False

        context['_trace'] = project['_transient']['trace']
        return True

    # Helpers

    def _get_mpdmanager(self) -> ManagedProjectDefinitionManager:
        if self._mpdmanager is None:
            self._mpdmanager = self.get_property('managedprojectdefinitions')
            if self._mpdmanager is None:
                raise ApiError('ManagedProjectDefinitionManager is None.')
            self._mpdmanager.platform = self
        return self._mpdmanager

    def _get_mamanager(self) -> ManagedAccountManager:
        if self._mamanager is None:
            self._mamanager = self.get_property('managedaccounts')
            if self._mpdmanager is None:
                raise ApiError('ManagedAccountManager is None.')
            self._mamanager.platform = self
        return self._mamanager



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
This module provides a basic _ManagedAccountManager_ helper managing
collections of managed accounts.
"""

from typing import Any, Dict

from zabel.commons.interfaces import ManagedAccount, Manager


########################################################################
########################################################################


class ManagedAccountManager(Manager):
    """Abstract ManagedAccount manager.

    Provides a default implementation for the set of features a managed
    account manager implementation must provide:

    - constructor (`__init__`)
    - ...

    The constructor, if overwritten, must take a metadata definition as
    a parameter, which is expected to be a dictionary with at least the
    following entries:

    - `storage (dict)`: a dictionary
    - `instances (class)`: a class

    `storage` is expected to contain a `type` entry (a class inheriting
    _CollectionStorage_).  It may contain other entries, as required by
    the said storage class.

    `instances` is a class inheriting _ManagedAccount_.
    """

    def __init__(self, configuration: Dict[str, Any]) -> None:
        self.configuration = configuration
        self._storage = configuration['storage']['type'](configuration)

    def get_managedaccount(self, canonical_id: str) -> ManagedAccount:
        raise NotImplementedError

    def create_managedaccount(
        self, canonical_id: str, account: ManagedAccount
    ) -> None:
        raise NotImplementedError

    def update_managedaccount(
        self, canonical_id: str, account: ManagedAccount
    ) -> None:
        raise NotImplementedError



# Copyright (c) 2020 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

from setuptools import setup, find_namespace_packages


with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='zabel',
    version='0.10.0',
    description='Zabel',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/engie-group/zabel',
    author='Martin Lafaix',
    author_email='martin.lafaix@external.engie.com',
    license='Eclipse Public License 2.0',
    packages=find_namespace_packages(include=['zabel.*']),
    package_data={'': ['*.yaml']},
    install_requires=[
        'requests>=2.23',
        'zabel-commons>=1.0',
        'zabel-fabric>=0.12',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='>= 3.6.5',
)



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
This module provides the _ManagedProjecDefinitionManager_ class.
"""


from typing import Any, Dict, List

from zabel.commons.interfaces import Manager, ManagedProjectDefinition
from zabel.commons.utils import ensure_instance, ensure_nonemptystring

########################################################################
## Abstract ManagedProjecDefinitionManager class


class ManagedProjectDefinitionManager(Manager):
    """Abstract ManagedProjectDefinition manager.

    Provides a default implementation for the set of features a managed
    project definition manager implementation must provide:

    - constructor (`__init__`)
    - `list_managedprojects`
    - `get_managedproject`
    - `create_managedproject`
    - `update_managedproject`

    Concrete classes deriving this abstract class must implement the
    following methods if they want to use the provided defaults:

    - `get_managedproject_key`
    - `get_key_managedproject`
    - `is_managedproject_key`

    The constructor, if overwritten, must take a metadata definition as
    a parameter, which is expected to be a dictionary with at least the
    following entries:

    - `storage (dict)`: a dictionary
    - `instances (class)`: a class

    `storage` is expected to contain a `type` entry (a class inheriting
    _CollectionStorage_).  It may contain other entries, as required by
    the said storage class.

    `instances` is a class inheriting _ManagedProjectDefinition_.
    """

    def __init__(self, configuration: Dict[str, Any]) -> None:
        self.configuration = configuration
        self._storage = configuration['storage']['type'](configuration)

    # abstract methods

    def get_managedproject_key(self, project_id: str) -> str:
        """Return the storage key for project_id.

        This method converts a project_id to the corresponding storage
        key.

        # Required parameters

        - `project_id`: a non-empty string

        # Returned value

        A non-empty string.
        """
        raise NotImplementedError

    def get_key_managedproject(self, key: str) -> str:
        """Return the project_id for storage key.

        This method converts a storage key to the corresponding
        project_id.

        It is only called on keys that are managed project definition
        keys.

        # Required parameters

        - `key`: a non-empty string

        # Returned value

        A non-empty string.
        """
        raise NotImplementedError

    def is_managedproject_key(self, key: str) -> bool:
        """Return True if key is a managed project definition key.

        The underlying storage collection may contain objects that are
        not managed project definitions.  This method is used to
        differentiate those.

        # Required parameters

        - `key`: a non-empty string

        # Returned value

        A boolean
        """
        raise NotImplementedError

    # default implementation

    def list_managedprojects(self) -> List[Dict[str, Any]]:
        """Return list of managed projects on platform.

        # Returned value

        A list.  Each item in the list is a dictionary with the
        following entries:

        - `project_id`: a string
        - `last_modified`: a timestamp or None
        """
        return [
            {
                'project_id': self.get_key_managedproject(item['key']),
                'last_modified': item.get('last_modified', None),
            }
            for item in self._storage.list()
            if self.is_managedproject_key(item['key'])
        ]

    def get_managedproject(self, project_id: str) -> ManagedProjectDefinition:
        """Return managed project details.

        # Required parameters

        - `project_id`: a non-empty string

        # Returned value

        Please refer to the concrete managed project definition used for
        more information on the returned value format.
        """
        ensure_nonemptystring('project_id')

        obj = self._storage.read(self.get_managedproject_key(project_id))
        mpd: ManagedProjectDefinition = self.configuration['instances']
        return mpd.from_dict(obj)

    def create_managedproject(
        self, project_id: str, project: ManagedProjectDefinition
    ) -> None:
        """Create new managed project.

        # Required parameters

        - `project_id`: a non-empty string
        - `project`: a dictionary

        `project_id` must not be the ID of an existing managed project.
        If it is, an _ApiError_ exception will be raised.
        """
        ensure_nonemptystring('project_id')
        ensure_instance('project', dict)

        self._storage.create(self.get_managedproject_key(project_id), project)

    def update_managedproject(
        self, project_id: str, project: ManagedProjectDefinition
    ) -> None:
        """Update managed project definition.

        # Required parameters

        - `project_id`: a non-empty string
        - `project`: a dictionary

        If no existing managed project with the provided ID exists,
        an _ApiError_ exception is raised.
        """
        ensure_nonemptystring('project_id')
        ensure_instance('project', dict)

        self._storage.update(self.get_managedproject_key(project_id), project)
